@@ -361,7 +361,7 @@ hnd_get(coap_context_t *ctx UNUSED_PARAM,
   coap_str_const_t *uri_path;
   int i;
   dynamic_resource_t *resource_entry = NULL;
-
+  coap_str_const_t value = { 0, NULL };
   /*
    * request will be NULL if an Observe triggered request, so the uri_path,
    * if needed, must be abstracted from the resource.
@@ -386,10 +386,14 @@ hnd_get(coap_context_t *ctx UNUSED_PARAM,
 
   resource_entry = &dynamic_entry[i];
 
+  if (resource_entry->value) {
+    value.length = resource_entry->value->length;
+    value.s = resource_entry->value->s;
+  }
   coap_add_data_blocked_response(resource, session, request, response, token,
                                  resource_entry->media_type, -1,
-                                 resource_entry->value->length,
-                                 resource_entry->value->s);
+                                 value.length,
+                                 value.s);
   return;
 }
 
@@ -436,6 +440,7 @@ hnd_put(coap_context_t *ctx UNUSED_PARAM,
     if (dynamic_count >= support_dynamic) {
       /* Should have been caught in hnd_unknown_put() */
       response->code = COAP_RESPONSE_CODE(406);
+      coap_delete_string(uri_path);
       return;
     }
     dynamic_count++;
@@ -462,7 +467,7 @@ hnd_put(coap_context_t *ctx UNUSED_PARAM,
       buf[sizeof(buf) - 1] = '\0';
       coap_add_attr(resource,
                     coap_make_str_const("ct"),
-                    coap_make_str_const(buf),
+                    coap_make_str_const((char*)buf),
                     0);
     } else {
       dynamic_count--;
@@ -474,6 +479,7 @@ hnd_put(coap_context_t *ctx UNUSED_PARAM,
     coap_delete_string(uri_path);
     response->code = COAP_RESPONSE_CODE(204);
     dynamic_entry[i].created = 0;
+    coap_resource_notify_observers(dynamic_entry[i].resource, NULL);
   }
 
   resource_entry = &dynamic_entry[i];
@@ -866,70 +872,6 @@ finish:
   return ctx;
 }
 
-static int
-join(coap_context_t *ctx, char *group_name){
-  struct ipv6_mreq mreq;
-  struct addrinfo   *reslocal = NULL, *resmulti = NULL, hints, *ainfo;
-  int result = -1;
-
-  /* we have to resolve the link-local interface to get the interface id */
-  memset(&hints, 0, sizeof(hints));
-  hints.ai_family = AF_INET6;
-  hints.ai_socktype = SOCK_DGRAM;
-
-  result = getaddrinfo("::", NULL, &hints, &reslocal);
-  if (result != 0) {
-    fprintf(stderr, "join: cannot resolve link-local interface: %s\n",
-            gai_strerror(result));
-    goto finish;
-  }
-
-  /* get the first suitable interface identifier */
-  for (ainfo = reslocal; ainfo != NULL; ainfo = ainfo->ai_next) {
-    if (ainfo->ai_family == AF_INET6) {
-      mreq.ipv6mr_interface =
-                ((struct sockaddr_in6 *)ainfo->ai_addr)->sin6_scope_id;
-      break;
-    }
-  }
-
-  memset(&hints, 0, sizeof(hints));
-  hints.ai_family = AF_INET6;
-  hints.ai_socktype = SOCK_DGRAM;
-
-  /* resolve the multicast group address */
-  result = getaddrinfo(group_name, NULL, &hints, &resmulti);
-
-  if (result != 0) {
-    fprintf(stderr, "join: cannot resolve multicast address: %s\n",
-            gai_strerror(result));
-    goto finish;
-  }
-
-  for (ainfo = resmulti; ainfo != NULL; ainfo = ainfo->ai_next) {
-    if (ainfo->ai_family == AF_INET6) {
-      mreq.ipv6mr_multiaddr =
-                ((struct sockaddr_in6 *)ainfo->ai_addr)->sin6_addr;
-      break;
-    }
-  }
-
-  if (ctx->endpoint) {
-    result = setsockopt(ctx->endpoint->sock.fd, IPPROTO_IPV6, IPV6_JOIN_GROUP, (char *)&mreq, sizeof(mreq));
-    if ( result == COAP_SOCKET_ERROR ) {
-      fprintf( stderr, "join: setsockopt: %s\n", coap_socket_strerror() );
-    }
-  } else {
-    result = -1;
-  }
-
- finish:
-  freeaddrinfo(resmulti);
-  freeaddrinfo(reslocal);
-
-  return result;
-}
-
 static ssize_t
 cmdline_read_key(char *arg, unsigned char *buf, size_t maxlen) {
   size_t len = strnlen(arg, maxlen);
@@ -951,7 +893,12 @@ main(int argc, char **argv) {
   coap_log_t log_level = LOG_WARNING;
   unsigned wait_ms;
   time_t t_last = 0;
+  int coap_fd;
+  fd_set m_readfds;
+  int nfds = 0;
+#ifndef _WIN32
   struct sigaction sa;
+#endif
 
   clock_offset = time(NULL);
 
@@ -1028,19 +975,59 @@ main(int argc, char **argv) {
 
   /* join multicast group if requested at command line */
   if (group)
-    join(ctx, group);
+    coap_join_mcast_group(ctx, group);
 
+  coap_fd = coap_context_get_coap_fd(ctx);
+  if (coap_fd != -1) {
+    /* if coap_fd is -1, then epoll is not supported within libcoap */
+    FD_ZERO(&m_readfds);
+    FD_SET(coap_fd, &m_readfds);
+    nfds = coap_fd + 1;
+  }
+
+#ifdef _WIN32
+  signal(SIGINT, handle_sigint);
+#else
   memset (&sa, 0, sizeof(sa));
   sigemptyset(&sa.sa_mask);
   sa.sa_handler = handle_sigint;
   sa.sa_flags = 0;
   sigaction (SIGINT, &sa, NULL);
   sigaction (SIGTERM, &sa, NULL);
+  /* So we do not exit on a SIGPIPE */
+  sa.sa_handler = SIG_IGN;
+  sigaction (SIGPIPE, &sa, NULL);
+#endif
 
   wait_ms = COAP_RESOURCE_CHECK_TIME * 1000;
 
   while ( !quit ) {
-    int result = coap_run_once( ctx, wait_ms );
+    int result;
+
+    if (coap_fd != -1) {
+      fd_set readfds = m_readfds;
+      struct timeval tv;
+
+      tv.tv_sec = wait_ms / 1000;
+      tv.tv_usec = (wait_ms % 1000) * 1000;
+      /* Wait until any i/o takes place */
+      result = select (nfds, &readfds, NULL, NULL, &tv);
+      if (result == -1) {
+        if (errno != EAGAIN) {
+          coap_log(LOG_DEBUG, "select: %s (%d)\n", coap_socket_strerror(), errno);
+          break;
+        }
+      }
+      if (result > 0) {
+        if (FD_ISSET(coap_fd, &readfds)) {
+          result = coap_run_once(ctx, COAP_RUN_NONBLOCK);
+        }
+      }
+    }
+    else {
+      /* epoll is not supported within libcoap */
+      result = coap_run_once( ctx, wait_ms );
+    }
     if ( result < 0 ) {
       break;
     } else if ( result && (unsigned)result < wait_ms ) {
@@ -1056,14 +1043,10 @@ main(int argc, char **argv) {
       time_t t_now = time(NULL);
       if (t_last != t_now) {
         /* Happens once per second */
-        int i;
         t_last = t_now;
         if (time_resource) {
           coap_resource_notify_observers(time_resource, NULL);
         }
-        for (i = 0; i < dynamic_count; i++) {
-          coap_resource_notify_observers(dynamic_entry[i].resource, NULL);
-        }
       }
       if (result) {
         /* result must have been >= wait_ms, so reset wait_ms */
