@@ -6,16 +6,12 @@
  * README for terms of use.
  */
 
-#include "coap_config.h"
+#include "coap_internal.h"
 
 #if defined(HAVE_STRNLEN) && defined(__GNUC__) && !defined(_GNU_SOURCE)
 #define _GNU_SOURCE 1
 #endif
 
-#if defined(HAVE_ASSERT_H) && !defined(assert)
-# include <assert.h>
-#endif
-
 #include <stdarg.h>
 #include <stdio.h>
 #include <string.h>
@@ -32,12 +28,6 @@
 #include <time.h>
 #endif
 
-#include "libcoap.h"
-#include "block.h"
-#include "coap_debug.h"
-#include "encode.h"
-#include "net.h"
-
 #ifdef WITH_LWIP
 # define fprintf(fd, ...) LWIP_PLATFORM_DIAG((__VA_ARGS__))
 # define fflush(...)
@@ -79,7 +69,7 @@ coap_set_log_level(coap_log_t level) {
 
 /* this array has the same order as the type log_t */
 static const char *loglevels[] = {
-  "EMRG", "ALRT", "CRIT", "ERR ", "WARN", "NOTE", "INFO", "DEBG"
+  "EMRG", "ALRT", "CRIT", "ERR ", "WARN", "NOTE", "INFO", "DEBG", "????", "CIPH"
 };
 
 #ifdef HAVE_TIME_H
@@ -173,11 +163,13 @@ coap_print_addr(const struct coap_address_t *addr, unsigned char *buf, size_t le
   const void *addrptr = NULL;
   in_port_t port;
   unsigned char *p = buf;
+  size_t need_buf;
 
   switch (addr->addr.sa.sa_family) {
   case AF_INET:
     addrptr = &addr->addr.sin.sin_addr;
     port = ntohs(addr->addr.sin.sin_port);
+    need_buf = INET_ADDRSTRLEN;
     break;
   case AF_INET6:
     if (len < 7) /* do not proceed if buffer is even too short for [::]:0 */
@@ -187,6 +179,7 @@ coap_print_addr(const struct coap_address_t *addr, unsigned char *buf, size_t le
 
     addrptr = &addr->addr.sin6.sin6_addr;
     port = ntohs(addr->addr.sin6.sin6_port);
+    need_buf = INET6_ADDRSTRLEN;
 
     break;
   default:
@@ -195,7 +188,8 @@ coap_print_addr(const struct coap_address_t *addr, unsigned char *buf, size_t le
   }
 
   /* Cast needed for Windows, since it doesn't have the correct API signature. */
-  if (inet_ntop(addr->addr.sa.sa_family, addrptr, (char *)p, len) == 0) {
+  if (inet_ntop(addr->addr.sa.sa_family, addrptr, (char *)p,
+                min(len, need_buf)) == 0) {
     perror("coap_print_addr");
     return 0;
   }
@@ -259,13 +253,13 @@ coap_print_addr(const struct coap_address_t *addr, unsigned char *buf, size_t le
 }
 
 #ifdef WITH_CONTIKI
-# define fprintf(fd, ...) PRINTF(__VA_ARGS__)
+# define fprintf(fd, ...) { (void)fd; PRINTF(__VA_ARGS__); }
 # define fflush(...)
 
 # ifdef HAVE_VPRINTF
-#  define vfprintf(fd, ...) vprintf(__VA_ARGS__)
+#  define vfprintf(fd, ...) { (void)fd; vprintf(__VA_ARGS__); }
 # else /* HAVE_VPRINTF */
-#  define vfprintf(fd, ...) PRINTF(__VA_ARGS__)
+#  define vfprintf(fd, ...) { (void)fd; PRINTF(__VA_ARGS__); }
 # endif /* HAVE_VPRINTF */
 #endif /* WITH_CONTIKI */
 
@@ -291,7 +285,7 @@ msg_code_string(uint16_t c) {
   } else if (c >= 224 && c - 224 < (int)(sizeof(signals)/sizeof(const char *))) {
     return signals[c-224];
   } else {
-    snprintf(buf, sizeof(buf), "%u.%02u", c >> 5, c & 0x1f);
+    snprintf(buf, sizeof(buf), "%u.%02u", (c >> 5) & 0x7, c & 0x1f);
     return buf;
   }
 }
@@ -459,7 +453,14 @@ is_binary(int content_format) {
 
 void
 coap_show_pdu(coap_log_t level, const coap_pdu_t *pdu) {
-  unsigned char buf[1024]; /* need some space for output creation */
+#if COAP_CONSTRAINED_STACK
+  static coap_mutex_t static_show_pdu_mutex = COAP_MUTEX_INITIALIZER;
+  static unsigned char buf[min(COAP_DEBUG_BUF_SIZE, 1024)]; /* need some space for output creation */
+  static char outbuf[COAP_DEBUG_BUF_SIZE];
+#else /* ! COAP_CONSTRAINED_STACK */
+  unsigned char buf[min(COAP_DEBUG_BUF_SIZE, 1024)]; /* need some space for output creation */
+  char outbuf[COAP_DEBUG_BUF_SIZE];
+#endif /* ! COAP_CONSTRAINED_STACK */
   size_t buf_len = 0; /* takes the number of bytes written to buf */
   int encode = 0, have_options = 0, i;
   coap_opt_iterator_t opt_iter;
@@ -467,13 +468,16 @@ coap_show_pdu(coap_log_t level, const coap_pdu_t *pdu) {
   int content_format = -1;
   size_t data_len;
   unsigned char *data;
-  char outbuf[COAP_DEBUG_BUF_SIZE];
   int outbuflen = 0;
 
   /* Save time if not needed */
   if (level > coap_get_log_level())
     return;
 
+#if COAP_CONSTRAINED_STACK
+  coap_mutex_lock(&static_show_pdu_mutex);
+#endif /* COAP_CONSTRAINED_STACK */
+
   snprintf(outbuf, sizeof(outbuf), "v:%d t:%s c:%s i:%04x {",
           COAP_DEFAULT_VERSION, msg_type_string(pdu->type),
           msg_code_string(pdu->code), pdu->tid);
@@ -637,8 +641,11 @@ coap_show_pdu(coap_log_t level, const coap_pdu_t *pdu) {
       snprintf(&outbuf[outbuflen], sizeof(outbuf)-outbuflen,  ">>");
     } else {
       if (print_readable(data, data_len, buf, sizeof(buf), 0)) {
+        size_t max_length;
         outbuflen = strlen(outbuf);
-        snprintf(&outbuf[outbuflen], sizeof(outbuf)-outbuflen,  "'%s'", buf);
+        max_length = sizeof(outbuf)-outbuflen;
+        if (snprintf(&outbuf[outbuflen], max_length,  "'%s'", buf) >= (int)max_length)
+          outbuf[sizeof(outbuf)-1] = '\000';
       }
     }
   }
@@ -646,6 +653,10 @@ coap_show_pdu(coap_log_t level, const coap_pdu_t *pdu) {
   outbuflen = strlen(outbuf);
   snprintf(&outbuf[outbuflen], sizeof(outbuf)-outbuflen,  "\n");
   COAP_DO_SHOW_OUTPUT_LINE;
+
+#if COAP_CONSTRAINED_STACK
+  coap_mutex_unlock(&static_show_pdu_mutex);
+#endif /* COAP_CONSTRAINED_STACK */
 }
 
 void coap_show_tls_version(coap_log_t level)
@@ -750,16 +761,24 @@ coap_log_impl(coap_log_t level, const char *format, ...) {
     return;
 
   if (log_handler) {
-#if defined(WITH_CONTIKI) || defined(WITH_LWIP)
-    char message[128];
-#else
-    char message[8 + 1024 * 2]; /* O/H + Max packet payload size * 2 */
-#endif
+#if COAP_CONSTRAINED_STACK
+    static coap_mutex_t static_log_mutex = COAP_MUTEX_INITIALIZER;
+    static char message[COAP_DEBUG_BUF_SIZE];
+#else /* ! COAP_CONSTRAINED_STACK */
+    char message[COAP_DEBUG_BUF_SIZE];
+#endif /* ! COAP_CONSTRAINED_STACK */
     va_list ap;
     va_start(ap, format);
+#if COAP_CONSTRAINED_STACK
+  coap_mutex_lock(&static_log_mutex);
+#endif /* COAP_CONSTRAINED_STACK */
+
     vsnprintf( message, sizeof(message), format, ap);
     va_end(ap);
     log_handler(level, message);
+#if COAP_CONSTRAINED_STACK
+    coap_mutex_unlock(&static_log_mutex);
+#endif /* COAP_CONSTRAINED_STACK */
   } else {
     char timebuf[32];
     coap_tick_t now;
@@ -772,7 +791,7 @@ coap_log_impl(coap_log_t level, const char *format, ...) {
     if (print_timestamp(timebuf,sizeof(timebuf), now))
       fprintf(log_fd, "%s ", timebuf);
 
-    if (level <= LOG_DEBUG)
+    if (level <= COAP_LOG_CIPHERS)
       fprintf(log_fd, "%s ", loglevels[level]);
 
     va_start(ap, format);
