@@ -19,16 +19,12 @@
  * g_env      A coap_gnutls_env_t * (held in c_session->tls)
  */
 
-#include "coap_config.h"
+#include "coap_internal.h"
 
 #ifdef HAVE_LIBGNUTLS
 
 #define MIN_GNUTLS_VERSION "3.3.0"
 
-#include "net.h"
-#include "mem.h"
-#include "coap_debug.h"
-#include "prng.h"
 #include <inttypes.h>
 #include <stdio.h>
 #include <errno.h>
@@ -64,6 +60,7 @@ typedef struct coap_gnutls_env_t {
   /* If not set, need to do gnutls_handshake */
   int established;
   int seen_client_hello;
+  int doing_dtls_timeout;
 } coap_gnutls_env_t;
 
 #define IS_PSK (1 << 0)
@@ -404,7 +401,7 @@ psk_client_callback(gnutls_session_t g_session,
   *username = gnutls_malloc(identity_len+1);
   if (*username) {
     memcpy(*username, identity, identity_len);
-    *username[identity_len] = '\0';
+    (*username)[identity_len] = '\0';
   }
 
   key->data = gnutls_malloc(psk_len);
@@ -1123,13 +1120,22 @@ receive_timeout(gnutls_transport_ptr_t context, unsigned int ms UNUSED) {
     fd_set readfds, writefds, exceptfds;
     struct timeval tv;
     int nfds = c_session->sock.fd +1;
+    coap_gnutls_env_t *g_env = (coap_gnutls_env_t *)c_session->tls;
+
+    /* If data has been read in by libcoap ahead of GnuTLS, say it is there */
+    if (c_session->proto == COAP_PROTO_DTLS && g_env &&
+        g_env->coap_ssl_data.pdu_len > 0) {
+      return 1;
+    }
 
     FD_ZERO(&readfds);
     FD_ZERO(&writefds);
     FD_ZERO(&exceptfds);
     FD_SET (c_session->sock.fd, &readfds);
-    FD_SET (c_session->sock.fd, &writefds);
-    FD_SET (c_session->sock.fd, &exceptfds);
+    if (!(g_env && g_env->doing_dtls_timeout)) {
+      FD_SET (c_session->sock.fd, &writefds);
+      FD_SET (c_session->sock.fd, &exceptfds);
+    }
     /* Polling */
     tv.tv_sec = 0;
     tv.tv_usec = 0;
@@ -1178,6 +1184,7 @@ coap_dtls_new_gnutls_env(coap_session_t *c_session, int type)
           "gnutls_priority_set");
   gnutls_handshake_set_timeout(g_env->g_session,
                                GNUTLS_DEFAULT_HANDSHAKE_TIMEOUT);
+  gnutls_dtls_set_timeouts(g_env->g_session, 5000, 60000);
 
   return g_env;
 
@@ -1213,7 +1220,9 @@ coap_dtls_free_gnutls_env(coap_gnutls_context_t *g_context,
         g_env->psk_sv_credentials = NULL;
       }
     }
-    if (g_context->psk_pki_enabled & IS_PKI) {
+    if ((g_context->psk_pki_enabled & IS_PKI) ||
+        (g_context->psk_pki_enabled &
+         (IS_PSK | IS_PKI | IS_CLIENT)) == IS_CLIENT) {
       gnutls_certificate_free_credentials(g_env->pki_credentials);
       g_env->pki_credentials = NULL;
     }
@@ -1223,11 +1232,9 @@ coap_dtls_free_gnutls_env(coap_gnutls_context_t *g_context,
 
 void *coap_dtls_new_server_session(coap_session_t *c_session) {
   coap_gnutls_env_t *g_env =
-         (coap_gnutls_env_t *)c_session->endpoint->hello.tls;
+         (coap_gnutls_env_t *)c_session->tls;
 
   gnutls_transport_set_ptr(g_env->g_session, c_session);
-  /* For the next one */
-  c_session->endpoint->hello.tls = NULL;
 
   return g_env;
 }
@@ -1340,6 +1347,7 @@ void coap_dtls_free_session(coap_session_t *c_session) {
                 COAP_PROTO_NOT_RELIABLE(c_session->proto) ?
                  COAP_FREE_BYE_AS_UDP : COAP_FREE_BYE_AS_TCP);
     c_session->tls = NULL;
+    coap_handle_event(c_session->context, COAP_EVENT_DTLS_CLOSED, c_session);
   }
 }
 
@@ -1411,18 +1419,40 @@ int coap_dtls_send(coap_session_t *c_session,
 }
 
 int coap_dtls_is_context_timeout(void) {
-  return 1;
+  return 0;
 }
 
 coap_tick_t coap_dtls_get_context_timeout(void *dtls_context UNUSED) {
   return 0;
 }
 
-coap_tick_t coap_dtls_get_timeout(coap_session_t *c_session UNUSED) {
+coap_tick_t coap_dtls_get_timeout(coap_session_t *c_session, coap_tick_t now) {
+  coap_gnutls_env_t *g_env = (coap_gnutls_env_t *)c_session->tls;
+
+  if (g_env && g_env->g_session) {
+    unsigned int rem_ms = gnutls_dtls_get_timeout(g_env->g_session);
+
+    return now + rem_ms;
+  }
+
   return 0;
 }
 
-void coap_dtls_handle_timeout(coap_session_t *c_session UNUSED) {
+void coap_dtls_handle_timeout(coap_session_t *c_session) {
+  coap_gnutls_env_t *g_env = (coap_gnutls_env_t *)c_session->tls;
+
+  assert(g_env != NULL);
+  g_env->doing_dtls_timeout = 1;
+  if (((c_session->state == COAP_SESSION_STATE_HANDSHAKE) &&
+       (++c_session->dtls_timeout_count > c_session->max_retransmit)) ||
+      (do_gnutls_handshake(c_session, g_env) < 0)) {
+    /* Too many retries */
+    g_env->doing_dtls_timeout = 0;
+    coap_session_disconnected(c_session, COAP_NACK_TLS_FAILED);
+  }
+  else {
+    g_env->doing_dtls_timeout = 0;
+  }
 }
 
 /*
@@ -1492,22 +1522,6 @@ coap_dtls_receive(coap_session_t *c_session,
   return ret;
 }
 
-#define DTLS_CT_HANDSHAKE          22
-#define DTLS_HT_CLIENT_HELLO        1
-
-/** Generic header structure of the DTLS record layer. */
-typedef struct __attribute__((__packed__)) {
-  uint8_t content_type;           /**< content type of the included message */
-  uint16_t version;               /**< Protocol version */
-  uint16_t epoch;                 /**< counter for cipher state changes */
-  uint8_t sequence_number[6];     /**< sequence number */
-  uint16_t length;                /**< length of the following fragment */
-  uint8_t handshake;              /**< If content_type == DTLS_CT_HANDSHAKE */
-} dtls_record_handshake_t;
-
-#define OFF_CONTENT_TYPE 0     /* offset of content_type in dtls_record_handshake_t */
-#define OFF_HANDSHAKE_TYPE 13  /* offset of handshake in dtls_record_handshake_t */
-
 /*
  * return 0 failed
  *        1 passed
@@ -1522,24 +1536,6 @@ coap_dtls_hello(coap_session_t *c_session,
   int ret;
 
   if (!g_env) {
-    /*
-     * Need to check that this actually is a Client Hello before wasting
-     * time allocating and then freeing off g_env.
-     */
-    if (data_len < (OFF_HANDSHAKE_TYPE + 1)) {
-      coap_log(LOG_DEBUG,
-         "coap_dtls_hello: ContentType %d Short Packet (%ld < %d) dropped\n",
-         data[OFF_CONTENT_TYPE], data_len, OFF_HANDSHAKE_TYPE + 1);
-      return 0;
-    }
-    if (data[OFF_CONTENT_TYPE] != DTLS_CT_HANDSHAKE ||
-        data[OFF_HANDSHAKE_TYPE] != DTLS_HT_CLIENT_HELLO) {
-      coap_log(LOG_DEBUG,
-         "coap_dtls_hello: ContentType %d Handshake %d dropped\n",
-         data[OFF_CONTENT_TYPE], data[OFF_HANDSHAKE_TYPE]);
-      return 0;
-    }
-
     g_env = coap_dtls_new_gnutls_env(c_session, GNUTLS_SERVER);
     if (g_env) {
       c_session->tls = g_env;
@@ -1605,6 +1601,13 @@ coap_sock_read(gnutls_transport_ptr_t context, void *out, size_t outl) {
 #else
     ret = recv(c_session->sock.fd, out, outl, 0);
 #endif
+    if (ret > 0) {
+      coap_log(LOG_DEBUG, "*  %s: setup: received %d bytes\n",
+               coap_session_str(c_session), ret);
+    } else if (ret < 0 && errno != EAGAIN) {
+      coap_log(LOG_DEBUG,  "*  %s: setup: failed to receive any bytes (%s)\n",
+               coap_session_str(c_session), coap_socket_strerror());
+    }
     if (ret == 0) {
       /* graceful shutdown */
       c_session->sock.flags &= ~COAP_SOCKET_CAN_READ;
@@ -1629,6 +1632,13 @@ coap_sock_write(gnutls_transport_ptr_t context, const void *in, size_t inl) {
   coap_session_t *c_session = (struct coap_session_t *)context;
 
   ret = (int)coap_socket_write(&c_session->sock, in, inl);
+  if (ret > 0) {
+    coap_log(LOG_DEBUG, "*  %s: setup: sent %d bytes\n",
+             coap_session_str(c_session), ret);
+  } else if (ret < 0) {
+    coap_log(LOG_DEBUG,  "*  %s: setup: failed to send %d bytes (%s)\n",
+             coap_session_str(c_session), ret, coap_socket_strerror());
+  }
   if (ret == 0) {
     errno = EAGAIN;
     ret = -1;
@@ -1772,7 +1782,9 @@ ssize_t coap_tls_write(coap_session_t *c_session,
   }
 
   if (c_session->dtls_event >= 0) {
-    coap_handle_event(c_session->context, c_session->dtls_event, c_session);
+    /* COAP_EVENT_DTLS_CLOSED event reported in coap_session_disconnected() */
+    if (c_session->dtls_event != COAP_EVENT_DTLS_CLOSED)
+      coap_handle_event(c_session->context, c_session->dtls_event, c_session);
     if (c_session->dtls_event == COAP_EVENT_DTLS_ERROR ||
         c_session->dtls_event == COAP_EVENT_DTLS_CLOSED) {
       coap_session_disconnected(c_session, COAP_NACK_TLS_FAILED);
@@ -1833,7 +1845,9 @@ ssize_t coap_tls_read(coap_session_t *c_session,
   }
 
   if (c_session->dtls_event >= 0) {
-    coap_handle_event(c_session->context, c_session->dtls_event, c_session);
+    /* COAP_EVENT_DTLS_CLOSED event reported in coap_session_disconnected() */
+    if (c_session->dtls_event != COAP_EVENT_DTLS_CLOSED)
+      coap_handle_event(c_session->context, c_session->dtls_event, c_session);
     if (c_session->dtls_event == COAP_EVENT_DTLS_ERROR ||
         c_session->dtls_event == COAP_EVENT_DTLS_CLOSED) {
       coap_session_disconnected(c_session, COAP_NACK_TLS_FAILED);
