@@ -6,7 +6,7 @@
  * README for terms of use.
  */
 
-#include "coap_config.h"
+#include "coap_internal.h"
 
 #ifdef HAVE_TIME_H
 #include <time.h>
@@ -21,9 +21,6 @@
 #include <stdint.h>
 #endif
 
-#include "libcoap.h"
-#include "coap_time.h"
-
 static coap_tick_t coap_clock_offset = 0;
 
 #if _POSIX_TIMERS && !defined(__APPLE__)
@@ -68,13 +65,13 @@ coap_clock_init(void) {
 }
 
 /* creates a Qx.frac from fval */
-#define Q(frac,fval) ((coap_tick_t)(((1 << (frac)) * (fval))))
+#define Q(frac,fval) ((1 << (frac)) * (fval))
 
 /* number of frac bits for sub-seconds */
 #define FRAC 10
 
 /* rounds val up and right shifts by frac positions */
-#define SHR_FP(val,frac) (((val) + (1 << ((frac) - 1))) >> (frac))
+#define SHR_FP(val,frac) (((coap_tick_t)((val) + (1 << ((frac) - 1)))) >> (frac))
 
 void
 coap_ticks(coap_tick_t *t) {
